﻿using System;
using System.IO;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.Windows.Forms;
using UsefulThings;

namespace KFreonLib.MEDirectories
{
    public static class ME3Directory
    {
        private static string _gamePath = null;
        public static string gamePath
        {
            get // if you are trying to use gamePath variable and it's null it asks to locate ME3 exe file
            {
                /*if (_gamePath == null)
                    _gamePath = KFreonLib.Misc.Methods.SelectGameLoc(3);
                else */
                if (_gamePath == null)
                    return null;

                if (!_gamePath.EndsWith("\\"))
                    _gamePath += '\\';
                return _gamePath;
            }
            set
            {
                if (value != null)
                {
                    if (value.Contains("BIOGame"))
                        value = value.Substring(0, value.LastIndexOf("BIOGame"));
                }
                _gamePath = value;
            }
        }
        public static string BIOGamePath { get { return (gamePath != null) ? gamePath.Contains("biogame", StringComparison.OrdinalIgnoreCase) ? gamePath : Path.Combine(gamePath, @"BIOGame\") : null; } }
        public static string tocFile { get { return (gamePath != null) ? Path.Combine(gamePath, @"BIOGame\PCConsoleTOC.bin") : null; } }
        public static string cookedPath { get { return (gamePath != null) ? Path.Combine(gamePath, @"BIOGame\CookedPCConsole\") : "Not Found"; } }
        public static string DLCPath { get { return (gamePath != null) ? Path.Combine(gamePath , @"BIOGame\DLC\") : "Not Found"; } }

        // "C:\...\MyDocuments\BioWare\Mass Effect 3\" folder
        public static string BioWareDocPath { get { return Environment.GetFolderPath(Environment.SpecialFolder.Personal) + @"\BioWare\Mass Effect 3\"; } }
        public static string GamerSettingsIniFile { get { return BioWareDocPath + @"BIOGame\Config\GamerSettings.ini"; } }

        public static string DLCFilePath(string DLCName)
        {
            string fullPath = DLCPath + DLCName + @"\CookedPCConsole\Default.sfar";
            if (File.Exists(fullPath))
                return fullPath;
            else
                throw new FileNotFoundException("Invalid DLC path " + fullPath);
        }

        static ME3Directory()
        {
            if (!string.IsNullOrEmpty(Properties.Settings.Default.ME3Directory))
            {
                gamePath = Properties.Settings.Default.ME3Directory;
            }
            else
            {
                string hkey32 = @"HKEY_LOCAL_MACHINE\SOFTWARE\";
                string hkey64 = @"HKEY_LOCAL_MACHINE\SOFTWARE\Wow6432Node\";
                string subkey = @"BioWare\Mass Effect 3";
                string keyName;

                keyName = hkey32 + subkey;
                string test = (string)Microsoft.Win32.Registry.GetValue(keyName, "Install Dir", null);
                if (test != null)
                {
                    gamePath = test;
                    return;
                }

                /*if (gamePath != null)
                    return;*/

                keyName = hkey64 + subkey;
                gamePath = (string)Microsoft.Win32.Registry.GetValue(keyName, "Install Dir", null); 
            }
        }

        static List<string> _files = null;
        public static List<string> Files
        {
            get
            {
                if (_files == null)
                {
                    Debugging.DebugOutput.PrintLn("ME3 COOKED: " + ME3Directory.cookedPath);
                    if (String.IsNullOrEmpty(ME3Directory.cookedPath) || !Directory.Exists(ME3Directory.cookedPath))
                        return null;

                    _files = MEDirectories.EnumerateGameFiles(3, ME3Directory.cookedPath);

                    Debugging.DebugOutput.PrintLn("ME3 DLC: " + ME3Directory.DLCPath);
                    if (!String.IsNullOrEmpty(ME3Directory.DLCPath) && Directory.Exists(DLCPath))
                        _files.AddRange(MEDirectories.EnumerateGameFiles(3, ME3Directory.DLCPath));
                }

                return _files;
            }
        }
    }
}
